package clustering.poboc;

import java.io.IOException;
import java.sql.SQLException;

import processutilities.core.GroupHandler;

import report.Timer;
import similarity.SimilarityFunction;
import similarity.matrix.AveragesTable;
import similarity.matrix.MatrixFiller;
import similarity.matrix.SimilaritiesTable;
import clustering.helpers.AddOverlappingClustersFilter;
import clustering.interfaces.ClusteringAlgorithm;
import clustering.interfaces.ProgressMonitor;
import clustering.poboc.sql.CliqueConstructor;
import clustering.poboc.sql.ClustersTable;
import clustering.poboc.sql.ConnectionsTable;
import clustering.poboc.sql.PolesTable;
import database.interfaces.DatabaseConnectionProvider;
import datastructures.interfaces.DataSet;

public class PoBOCAlgorithm implements
		ClusteringAlgorithm<PoBOCParametersContainer> {
	
	private boolean closeWhenFinish;
	private SimilaritiesTable similaritiesTable;
	private AveragesTable averagesTable;
	private ConnectionsTable connectionsTable;
	private CliqueConstructor cliqueConstructor;
	private PolesTable polesTable;
	private ClustersTable clustersTable;
	private Timer timer;
	
	public ClustersTable buildClusters(DataSet dataSet,
			DatabaseConnectionProvider database, double samplingFactor,
			SimilarityFunction similarityFunction, ProgressMonitor monitor) throws SQLException {

		monitor.init("PoBOCAlgorithm", 5);
						
		// Creo e riempio la tabella di similarità
		monitor.setMessage("Building Similarity Matrix");
		// check = 1
		checkPoint();
		MatrixFiller matrixFiller = new MatrixFiller(true,dataSet,database,similarityFunction);
		matrixFiller.setSamplingFactor(samplingFactor);
		float averageSimilarity = matrixFiller.fillMatrix(monitor);
		similaritiesTable = matrixFiller.getSimilaritiesTable();
		averagesTable = matrixFiller.getAveragesTable();
		
		System.out.println("Free memory: " + Runtime.getRuntime().freeMemory() / 1024);
		if(monitor.isCanceled())
			return null;
		monitor.tick();
		
		// Riempo la tabella delle connessioni
		monitor.setMessage("Building Similarity Graph");
		// check = 2
		checkPoint();
		connectionsTable= new ConnectionsTable(database.connect());
		connectionsTable.fillConnectionTable(similaritiesTable, averagesTable);

		System.runFinalization();
		System.gc();
		System.out.println("Free memory: " + Runtime.getRuntime().freeMemory() / 1024);
		
		if(monitor.isCanceled())
			return null;
		monitor.tick();
		
		// Trovo i poli (le cricche)
		monitor.setMessage("Searching for poles");
		// check = 3
		checkPoint();
		cliqueConstructor = new CliqueConstructor(database.connect(), similaritiesTable, averagesTable, connectionsTable);
		cliqueConstructor.buildCliques(averageSimilarity);
		polesTable = cliqueConstructor.getPoleTable();
		connectionsTable = cliqueConstructor.getConnTable();
		
		System.runFinalization();
		System.gc();
		System.out.println("Free memory: " + Runtime.getRuntime().freeMemory() / 1024);
		
		if(monitor.isCanceled())
			return null;
		monitor.tick();
		
		// Assegnamento ai cluster (A ogni polo corrisponde un cluster)
		monitor.setMessage("Assigning objects to clusters");
		// check = 4
		checkPoint();
		clustersTable = new ClustersTable(database.connect(), polesTable);
		clustersTable.doClustering(similaritiesTable);
		
		System.runFinalization();
		System.gc();
		System.out.println("Free memory: " + Runtime.getRuntime().freeMemory() / 1024);
		
		if(monitor.isCanceled())
			return null;
		monitor.tick();
			
		return clustersTable;
	}

	@Override
	public void doClustering(PoBOCParametersContainer parameter) {
		try {
			
			timer = new Timer();
			// check = 0
			timer.init();
			
			ClustersTable clustersTable = buildClusters(parameter.getDataSet(),
					parameter.getDataBase(), parameter.getSamplingFactor(), 
					parameter.getSimilarityFunction(),
					parameter.getProgressMonitor());
			
			if(clustersTable == null) {
				if(closeWhenFinish)
					close();
				return;
			}
			
			parameter.getProgressMonitor().setMessage("Building Output");
			// check = 5
			checkPoint();
			
			PoBOCResultFetcher fetcher = new PoBOCResultFetcher(clustersTable);
			GroupHandler.runAndJoin(parameter.getDataSet().buildInputHandler(), parameter.getOutputHandler(), 
					new AddOverlappingClustersFilter(fetcher));
			
			parameter.getProgressMonitor().done();
			
			// check = 6
			timer.end();
			
			this.doReport(parameter);
						
			if(closeWhenFinish)
				close();
		} catch (Exception e) {
			e.printStackTrace();
			try {
				close();
			} catch (SQLException e1) {
				e1.printStackTrace();
			}
			throw new RuntimeException(e);
		}
	}
	
	public void doReport(PoBOCParametersContainer parameter) throws SQLException, IOException {
		parameter.getReport().addTitleToReport("PoBOC Report");			
		doReportTimings(parameter);
		doReportSizes(parameter);
	}

	public void doReportSizes(PoBOCParametersContainer parameter)
			throws SQLException {
		parameter.getReport().addProperty("Sizes:");
		parameter.getReport().addProperty("Size of Similarities Matrix" , Integer.toString(similaritiesTable.getSize())+" tuples");
		parameter.getReport().addProperty("Size of Connection Table" , Integer.toString(connectionsTable.getSize())+" tuples");
		
		for(int i=0; i < polesTable.getNumberPolesFound(); i++) {
			parameter.getReport().addProperty("Size of Pole "+i , Integer.toString(polesTable.getNumberElementPole(i))+" objects");	
		}
		parameter.getReport().addProperty("Size of Poles Table" , Integer.toString(polesTable.getSize())+" tuples");
		
		for(int i=0; i < clustersTable.getNumberOfClustersFound(); i++) {
			parameter.getReport().addProperty("Size of Cluster "+i , Integer.toString(clustersTable.getNumberOfElementsInCluster(i))+" objects");
		}
		parameter.getReport().addProperty("Size of Cluster Table" , Integer.toString(clustersTable.getSize())+" tuples");
	}

	public void doReportTimings(PoBOCParametersContainer parameter) {
		parameter.getReport().addProperty("Timings:");
		parameter.getReport().addProperty("Time to build the Similarities Matrix", Long.toString(timer.getTimeForCheckPoint(1))+" sec");
		parameter.getReport().addProperty("Time to build the Connection Table", Long.toString(timer.getTimeForCheckPoint(2))+" sec");
		parameter.getReport().addProperty("Time to search the Poles", Long.toString(timer.getTimeForCheckPoint(3))+" sec");
		parameter.getReport().addProperty("Time to assign to Clusters", Long.toString(timer.getTimeForCheckPoint(4))+" sec");
		parameter.getReport().addProperty("Time to build the Output", Long.toString(timer.getTimeForCheckPoint(5))+" sec");
		parameter.getReport().addProperty("Overall execution time", Long.toString(timer.getTimeForAll())+" sec");
	}

	public void close() throws SQLException {
		if(similaritiesTable != null)
			similaritiesTable.close();
		if(averagesTable != null)
			averagesTable.close();
		if(connectionsTable != null)
			connectionsTable.close();
		if(cliqueConstructor != null)
			cliqueConstructor.close();
		if(polesTable != null)
			polesTable.close();
		if(clustersTable != null)
			clustersTable.close();
	}

	public SimilaritiesTable getSimilaritiesTable() {
		return similaritiesTable;
	}

	public AveragesTable getAveragesTable() {
		return averagesTable;
	}

	public ConnectionsTable getConnectionsTable() {
		return connectionsTable;
	}

	public PolesTable getPolesTable() {
		return polesTable;
	}

	public ClustersTable getClustersTable() {
		return clustersTable;
	}

	public boolean isCloseWhenFinish() {
		return closeWhenFinish;
	}

	public void setCloseWhenFinish(boolean closeWhenFinish) {
		this.closeWhenFinish = closeWhenFinish;
	}

	public void checkPoint() {
		if (timer == null) {
			timer= new Timer();
			timer.init();
		}
		timer.checkPoint();
	}

	public Timer getTimer() {
		return timer;
	}
}
