package clustering.poboc;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;

import processutilities.core.GroupHandler;
import similarity.SimilarityFunction;
import clustering.helpers.AddOverlappingClustersFilter;
import clustering.helpers.OverlappingClustersResultFetcher;
import clustering.interfaces.ClusteringAlgorithm;
import clustering.interfaces.ProgressMonitor;
import clustering.poboc.sql.ClustersTable;
import clustering.singlelink.SingleLinkAlgorithm;
import clustering.singlelink.datastructures.GraphPropertyBuilder;
import clustering.singlelink.datastructures.Group;
import clustering.singlelink.datastructures.GroupOfGroups;
import clustering.singlelink.datastructures.GroupOfTuples;
import clustering.singlelink.datastructures.GroupVisitor;
import database.core.HSQLDBConnectionProvider;
import database.interfaces.DatabaseConnectionProvider;
import datastructures.extra.MemoryDataSet;
import datastructures.interfaces.DataSet;
import datastructures.interfaces.NumericCell;
import datastructures.interfaces.Relation;
import datastructures.interfaces.Row;
import datastructures.simple.SimpleNumericAttribute;
import datastructures.simple.SimpleRelation;

public class PoBOCSingleLinkAlgorithm implements ClusteringAlgorithm<PoBOCParametersContainer> {

	private PoBOCAlgorithm poboc = new PoBOCAlgorithm();
	
	@Override
	public void doClustering(PoBOCParametersContainer parameter) {
		DatabaseConnectionProvider linksysDatabase = null;
		try {
			ClustersTable clustersTable = poboc.buildClusters(parameter.getDataSet(),
					parameter.getDataBase(), parameter.getSamplingFactor(), 
					parameter.getSimilarityFunction(),
					parameter.getProgressMonitor());
			
			if(clustersTable == null) {
				poboc.close();
				return;
			}
			
			parameter.getProgressMonitor().setMessage("Building Output");
			
			// Checkpoint 5
			poboc.checkPoint();
			
			int[] clusters = clustersTable.getClusters();
			String[][] csv = new String[clusters.length][1];
			for(int i=0; i < clusters.length; i++) {
				csv[i][0] = Integer.toString(clusters[i]);
			}
			
			SimpleRelation relation = new SimpleRelation(parameter.getDataSet().getRelation().getName());
			relation.addAttribute(new SimpleNumericAttribute("Cluster id"));
			
			MemoryDataSet dataSet = new MemoryDataSet(csv, relation);

			linksysDatabase = new HSQLDBConnectionProvider(
					"PoBOCSingleLinkAlgorithm");
			SingleLinkAlgorithm linksys = new SingleLinkAlgorithm();
			GroupOfGroups root = linksys
					.buildGrupTree(dataSet, linksysDatabase,
							new ClusterSimilarity(parameter.getDataBase()),
							new IgnoringProgressMonitor(parameter
									.getProgressMonitor()), clusters.length);

			if(root != null) {
				PoBOCSingleLinkResultFetcher fetcher = new PoBOCSingleLinkResultFetcher(
						root, clustersTable);
				GroupHandler.runAndJoin(parameter.getDataSet()
						.buildInputHandler(), parameter.getOutputHandler(),
						new AddOverlappingClustersFilter(fetcher));
				parameter.getProgressMonitor().done();
			}
			
			// checkpoint 6
			poboc.getTimer().end();
			
			doReport(parameter, root, csv.length);
			
			linksysDatabase.close();
			poboc.close();
		} catch (Exception e) {			
			e.printStackTrace();

			if(linksysDatabase != null)
				try {
					linksysDatabase.close();
				} catch (SQLException e2) {
					e2.printStackTrace();
				}
			
			try {				
				poboc.close();
			} catch (SQLException e1) {
				e1.printStackTrace();
			}
			throw new RuntimeException(e);
		}
		
	}
	
	private void doReport(PoBOCParametersContainer parameter, GroupOfGroups root, int size) throws SQLException, IOException {
		parameter.getReport().setTitle("PoBOC+SingleLinkage");
		poboc.doReportTimings(parameter);
		poboc.doReportSizes(parameter);
		
		parameter.getReport().addProperty("Groups:");
		
		GraphPropertyBuilder builder = new GraphPropertyBuilder();
		root.accept(builder);
		
		parameter.getReport().addProperty(builder.getGraphProperty());
	}

	private class ClusterSimilarity implements SimilarityFunction, Cloneable {

		private final DatabaseConnectionProvider database;
		
		private ClustersTable clustersTable;
		
		public ClusterSimilarity(DatabaseConnectionProvider database) {
			this.database = database;
		}
		
		@Override
		public boolean canBeAppliedTo(Relation relation) {
			if(relation.getAttributes().length == 1 && relation.getAttribute(0).getName().equals("Cluster Id"))
				return true;
			return false;
		}

		@Override
		public float compare(Row r1, Row r2) {
			int id1 = (int) ((NumericCell) r1
					.getCells()[0]).getValue();
			int id2 = (int) ((NumericCell) r2
					.getCells()[0]).getValue();
			//System.out.println("Comparing " + id1 + " and " + id2);
			try {
				return getClustersTable().getSimClusters(id1,id2);
			} catch (SQLException e) {
				throw new RuntimeException(e);
			}
		}

		@Override
		public void preprocess(DataSet dataSet) {			
		}

		@Override
		public void dispose() {
			try {
				if(clustersTable != null)
					clustersTable.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		
		@Override
		public SimilarityFunction duplicate() {
			ClusterSimilarity clone = null;
			try {
				clone = (ClusterSimilarity) super.clone();
				clone.clustersTable = null;
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();
			}
			return clone;			
		}

		private ClustersTable getClustersTable() throws SQLException {
			if(clustersTable == null)
				clustersTable = new ClustersTable(database.connect(),false);
			return clustersTable;
		}
		
	}
	
	private class PoBOCSingleLinkResultFetcher implements OverlappingClustersResultFetcher {

		private String[] clusters;
		private boolean[][] booleanTables;
		private GroupOfGroups root;
		private ClustersTable table;
		
		public PoBOCSingleLinkResultFetcher(GroupOfGroups root, ClustersTable table) {
			this.root = root;
			this.table = table;
		}
		
		@Override
		public String[] getClusters() {
			if(clusters == null) {
				ClustersBuilder builder;
				try {
					builder = new ClustersBuilder(table.getClusters());
				} catch (SQLException e) {
					e.printStackTrace();
					throw new RuntimeException();
				}
				root.accept(builder);
				clusters = builder.getClusters();
			}
			return clusters;
		}

		@Override
		public boolean[] getClustersArray(int rowId) {
			boolean[] result = new boolean[getClusters().length];
			
			boolean[][] tables = getBooleanTables();
			
			int[] clusters;
			try {
				clusters = table.getClusters(rowId);
			} catch (SQLException e) {
				e.printStackTrace();
				throw new RuntimeException();
			}
			
			for(int cluster : clusters) {
				for(int i=0; i < tables[cluster].length; i++) {
					if(tables[cluster][i])
						result[i] = true;
				}
			}
			
			return result;
		}
		
		private boolean[][] getBooleanTables() {
			if( booleanTables == null ){
//				String cl = "";
//				for(String s : getClusters()) {
//					cl += s;
//					cl += " ";
//				}
//				System.out.println(cl);
				int[] initialClusters;
				try {
					initialClusters = table.getClusters();
				} catch (SQLException e) {
					e.printStackTrace();
					throw new RuntimeException();
				}
				booleanTables = new boolean[initialClusters.length][getClusters().length];
				
				for(int k : initialClusters) {
					TableBuilder tb = new TableBuilder(initialClusters.length, getClusters().length, k);
					root.accept(tb);
					booleanTables[k] = tb.getTupleTable();
				}
			}
			return booleanTables;
		}
		
	}
	
	private class ClustersBuilder implements GroupVisitor {

		private ArrayList<String> clusters = new ArrayList<String>();
		private int i=0;
		
		public ClustersBuilder(int[] clusterIds) {
			Arrays.sort(clusterIds);
			for(int k: clusterIds) {
				clusters.add("c" + k);
			}
			i = clusterIds.length;
		}
		
		@Override
		public void visit(GroupOfTuples groupOfTuples) {
		}

		@Override
		public void visit(GroupOfGroups groupOfGroups) {
			for(Group group : groupOfGroups.getGroups()) {
				group.accept(this);
			}
			groupOfGroups.setId(i);
			clusters.add("g" + i);
			i++;
		}
		
		public String[] getClusters() {
			clusters.remove(clusters.size() -1);
			return clusters.toArray(new String[0]);
		}
		
	}
	
	private class TableBuilder implements GroupVisitor {

		private final boolean[] table;
		private final int offset;
		private final int tupleIndex;
		private int i = 0;
		
		private boolean childContainsTuple = false;
		
		public TableBuilder(int offset, int size, int tupleIndex) {
			table = new boolean[size];
			this.offset = offset;
			this.tupleIndex = tupleIndex;
		}
		
		@Override
		public void visit(GroupOfTuples groupOfTuples) {
			for(int k : groupOfTuples.getTuples()) {
				if(k == tupleIndex) {
					childContainsTuple = true;
					table[k] = true;
				}
			}
		}

		@Override
		public void visit(GroupOfGroups groupOfGroups) {
			boolean localChildContainsTuple = false;
			for(Group group : groupOfGroups.getGroups()) {
				group.accept(this);
				if(childContainsTuple)
					localChildContainsTuple = true;
				childContainsTuple = false;
			}
			childContainsTuple = localChildContainsTuple;
			if(offset + i < table.length) //if it doesn't happen we are the root.
				table[offset + i] = childContainsTuple;
			i++;
		}
		
		public boolean[] getTupleTable() {
			return table;
		}
		
	}
	
	private class IgnoringProgressMonitor implements ProgressMonitor {

		private final ProgressMonitor monitor;
		
		public IgnoringProgressMonitor(ProgressMonitor monitor) {
			this.monitor = monitor;
		}
		
		@Override
		public void done() {
		}

		@Override
		public int getMaximumTicks() {
			return 0;
		}

		@Override
		public String getMessage() {
			return null;
		}

		@Override
		public String getTaskName() {
			return null;
		}

		@Override
		public void init(String taskName, int maxTicks) {
			
		}

		@Override
		public boolean isCanceled() {
			return monitor.isCanceled();
		}

		@Override
		public void setCanceled(boolean canceled) {			
		}

		@Override
		public void setMessage(String message) {			
		}

		@Override
		public void tick() {			
		}
		
	}

}
