package clustering.poboc.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import similarity.matrix.AveragesTable;
import similarity.matrix.SimilaritiesTable;

/**
 * This class encapsulates the poles-building algorithm, i.e.,
 * the algorithm to build a clique and the algorithm to choose
 * the next centroid.
 * 
 * @author Luca Bernardini, Matteo Collina
 *
 */
public class CliqueConstructor {

	// QUERY SCELTA PRIMO VERTICE DI PARTENZA
	// Questa query verra invocata una sola volta nel processo di clustering perchè serve per scegliere il primo vertice 
	// per la costruzione della prima cricca
	//  Join tra Connections.Id_i e Averages.Id con ORDER BY su RowAvg e DISTINCT: Ottengo così tutti i valori di RowAvg dei 
	//	vertici con almeno una connessione (ordinati su RowAvg magari in ordine crescente dato che devo prendere 
	//  quello con minor RowAvg come centroide della prima cricca)
	/**
	 * This query is invokated only once in the clustering process because its used to choose
	 * the first vertex for the costruction of the first clique.
	 * It returns alle the ids that are connected ordering them by decreasing average similarity.
	 */
	public static final String queryOrderingAvgObjConnected = "SELECT C.Id_i "
			+ "FROM Connections C JOIN Averages A ON (C.Id_i = A.Id) "
			+ "ORDER BY A.RowAvg";

	// QUERY SCELTA ULTERIORI VERTICE DI PARTENZA
	// L'oggetto, tra quelli connessi, che ha la TotalPoleSim minima, ovvero l'oggetto più
	// lontano da tutti i poli, è il primo oggetto restituito da questa query.
	// Invoco questa query ogni qualvolta devo trovare un nuovo vertice di partenza per la ulteriore cricca da costruire.
	
	/**
	 * This query its used to find the connected object with the minimum total pole similarity,
	 * which is the further object from all the poles.
	 * 
	 * This query is invokated every time I need to find a new vertex to build a new clique.
	 * The nested query checks that the returned objects isn't connected with any object belonging to
	 * a clique. This condition is stronger than saying that the object shouldn't belong to a
	 * clique but it's needed because there might be problems during the best-in algorithm.
	 * In particular this avoids to find a clique of only one element. 
	 */
	public static final String queryPoleSim = "SELECT P.Id, AVG(P.PoleSim) AS TotalPoleSim "
			+ "FROM Nodes N JOIN PoleSimilarities P ON (N.Id = P.Id) "			
			// Query innestata che controlla che tra gli oggetti non vi siano anche quelli che già fanno parte di un polo (cricca)
			+ "WHERE NOT EXISTS (" +
					"SELECT PL.Id " +
					"FROM Poles PL " +
					"WHERE PL.Id = P.Id " +
					") "
			+ "GROUP BY P.Id "
			+ "ORDER BY TotalPoleSim";

	// BEST-IN : ALGORITMO DI COSTRUZIONE DELLA CRICCA
	// QUERY CHE RESTITUISCE TUTTI GLI ELEMENTI CONNESSI A UN ALTRO
	// Serve quando devo costruire una cricca: in questo caso il vertice di partenza è l'input della query, perchè
	// l'algoritmo Best-In prevede di trovare il nodo, tra quelli connessi a tutti i nodi della cricca in costruzione, 
	// che è più simile ad esso.
	// La query torna un ranking (dalla similarità più alta alla più bassa) degli oggetti che sono connessi
	// con tutti gli oggetti che già fanno parte della cricca in costruzione (prima sotto-query) ma che però
	// non fanno parte già di altre cricche (seconda sotto-query)
	/**
	 * This query is used when we need to build a clique from a starting vertex. It encapsulates the
	 * best-in algorithm, so it finds the first object that is connected to all the nodes already in the
	 * clique which is more similar to the starting vertex.
	 * It also discards all objects already belonging to a clique, i.e., a object can belong to only one 
	 * clique.
	 */
	public static final String queryGetBestIn = "SELECT C.Id_j "
			+ "FROM Connections C "
			+ "WHERE C.Id_i = ? "
			+ "AND ( "
				// QUERY INNESTATA: ritorna il numero di nodi con cui C.Id_j è connesso all'interno del polo
				// selezionato. Se questo numero è pari al numero totale di elementi nel polo, allora l'elemento
				// considerato può far parte della cricca.
				+ "SELECT COUNT(*) "
				+ "FROM Connections C2 JOIN Poles P ON (C2.Id_i = P.Id) "
				+ "WHERE C2.Id_j = C.Id_j AND P.Pole = ? "
				+ "GROUP BY P.Pole ) = ? "
			+ "AND NOT EXISTS (" +
				// QUERY INNESTATA: evita che tra gli oggetti tornati vi siano degli oggetti già appartenenti
				// ad altre cricche
					"SELECT P.Id " +
					"FROM Poles P " +
					"WHERE P.id = C.Id_j ) " +
			"ORDER BY C.Sim DESC";

	/** 
	 *  The connection through we interact with the database.
	 */
	private Connection connection = null;

	private ConnectionsTable connTable;
	private SimilaritiesTable simTable;
	private PolesTable poleTable;
	private AveragesTable avgTable;

	// Questi attributi sono invece le PreparateStatement che vengono invocate dai metodi che le incapsulano.
	// Grazie a questo approccio la costruzione della PreparateStatement avviene una volta sola (nel metodo get),
	// garantendo un risparmio evidente in sede di chiamata multipla delle query.
	private PreparedStatement psQueryPoleSim;
	private PreparedStatement psQueryGetBestIn;

	/**
	 * @param connection
	 * @param simTable
	 * @param avgTable
	 * @throws SQLException 
	 */
	public CliqueConstructor(Connection connection, SimilaritiesTable simTable, AveragesTable avgTable) throws SQLException {
		super();

		this.connection = connection;

		this.avgTable = avgTable;
		this.simTable = simTable;

		// Creo la tabella Poles e la vista PoleCounts
		this.poleTable = new PolesTable(connection);
		
		// Crea e riempo la tabella Connections e creo la vista Nodes
		this.connTable = new ConnectionsTable(connection);
		
		connTable.fillConnectionTable(this.simTable, this.avgTable);
	}
	
	/**
	 * @param connection
	 * @param simTable
	 * @param avgTable
	 * @param connTable
	 * @throws SQLException 
	 */
	public CliqueConstructor(Connection connection, SimilaritiesTable simTable, AveragesTable avgTable, ConnectionsTable connTable) throws SQLException {
		super();

		this.connection = connection;

		this.avgTable = avgTable;
		this.simTable = simTable;
		this.connTable = connTable;

		// Creo la tabella Poles e la vista PoleCounts
		this.poleTable = new PolesTable(connection);
	}
	
	/**
	 * This methods finds all the possibles cliques. It uses the bestIn method.
	 */
	public void buildCliques(float threshold) throws SQLException {
		poleTable.getConnection().setAutoCommit(false);
		
		// Inizialization first clique
		Statement stmt = connection.createStatement();
		stmt.setMaxRows(1);
		ResultSet firstCentroid = stmt.executeQuery(queryOrderingAvgObjConnected);
		
		if (!firstCentroid.next())  throw new RuntimeException("Dataset is empty");
		
		int poleUnderConstruction=0;
		int centroid = firstCentroid.getInt(1);
		firstCentroid.close();
		stmt.close();

		// Construction first clique and other cliques if threshold respect
		PreparedStatement ps= this.getPsPoleSim();
		ResultSet otherCentroid = null;
		//float threshold = this.avgTable.getTotalAverage();	// This is a query at AveragesTable
		float avg = 0;	// TotalPolSim of centroid
		int deleted = 0;	// numero di centroidi eliminati perchè non avevano connessioni
		boolean finish = false;
		
		while (avg < threshold) {
			System.out.println("Building pole: " + poleUnderConstruction
					+ ", centroid is: " + centroid);

			// System.out.println("Centroid: "+centroid+" AvgCentroid: "+avg+ "
			// threshold: "+threshold);
			poleTable.insertPole(centroid, poleUnderConstruction);
			int cliqueSize = bestIn(centroid, poleUnderConstruction);

			if (cliqueSize == 0) {
				System.out
						.println("This object has no connections: it isn't a Pole centroid ");
				poleTable.deletePole(poleUnderConstruction);
				deleted++;
			} else {
				System.out.println("Pole " + poleUnderConstruction
						+ " has been built: size = " + cliqueSize + " objects");
				poleUnderConstruction++;
			}

			poleTable.getConnection().commit();
			
			// ranking di oggetti ordinati su TotalPoleSim in modo crescente
			otherCentroid = ps.executeQuery();

			// in questo modo scorro il result set fino a quell'oggetto
			// che non ho ancora considerato, evitando di avere di nuovo come
			// centroide un oggetto scartato (deleted) in precedenza
			for (int j = 0; j <= deleted; j++) {
				if (!otherCentroid.next())
					finish = true;
			}
			if (!finish) {
				centroid = otherCentroid.getInt(1); // il primo P.Id è
													// l'elemento con
													// TotalPolSim minima
				avg = otherCentroid.getFloat(2);
				otherCentroid.close();
			} else
				break;
			System.out.println("avg = " + avg + ", threshold = " + threshold );
			
			System.runFinalization();
			System.gc();
		}
		
		//avgTable.printTable();
		poleTable.getConnection().commit();
		poleTable.getConnection().setAutoCommit(true);
	}

	// Ritorno un intero che è il numero di oggetti (escluso il centroide già inserito) che compongono 
	// la cricca che ho appena costruito. Se ritorna 0 vuol dire che non ho una cricca e dovrò eliminare il centroide
	// dalla tabella Poles prima di continuare (le cricche infatti devono avere almeno 2 oggetti connessi)
	/** 
	 * This method build the clique starting from input centroid. 
	 * The clique is built directly in the Poles table which is stored in th database, and not in memory.
	 */	
	protected int bestIn(int centroid, int poleUnderConstruction)
			throws SQLException {
		int poleCount=1;
		PreparedStatement ps = this.getPsGetBestIn();
		ps.setMaxRows(1);
		ps.setInt(1, centroid);
		ps.setInt(2, poleUnderConstruction);
		ps.setInt(3, poleCount);
		ResultSet connectedOrderedObject = ps.executeQuery();	
		// Nel result set ci sono tutti gli oggetti connessi a centroid in ordine dal più simile al meno simile
		while (connectedOrderedObject.next()) {
			if((poleCount % 10) == 0)
				System.out.println("added " + poleCount + " objects to pole " + poleUnderConstruction );
			poleCount++;
			poleTable.insertPole(connectedOrderedObject.getInt(1), poleUnderConstruction); // il primo è il più simile al centroide
			connectedOrderedObject.close();
			ps.setInt(3, poleCount);
			connectedOrderedObject = ps.executeQuery();	// cerco il prossimo elemento da aggiungere alla cricca rieseguendo la query
		}
		connectedOrderedObject.close();
		return poleCount - 1;		
	}

	//
	//  GET delle PreparedStatement: quando le invoco, se non è ancora stata creata, 
	//  viene creata la PreparedStatment passando la stringa relativa
	//

	private PreparedStatement getPsPoleSim() throws SQLException {
		if (psQueryPoleSim == null)
			psQueryPoleSim = connection.prepareStatement(CliqueConstructor.queryPoleSim);
		return psQueryPoleSim;
	}

	private PreparedStatement getPsGetBestIn() throws SQLException {
		if (psQueryGetBestIn == null)
			psQueryGetBestIn = connection.prepareStatement(CliqueConstructor.queryGetBestIn);
		return psQueryGetBestIn;
	}
	
	// GET TABLE
	
	/**
	 * @return the poleTable
	 */
	public PolesTable getPoleTable() {
		return poleTable;
	}

	/**
	 * @return the avgTable
	 */
	public AveragesTable getAvgTable() {
		return avgTable;
	}

	/**
	 * @return the connTable
	 */
	public ConnectionsTable getConnTable() {
		return connTable;
	}

	/**
	 * @return the simTable
	 */
	public SimilaritiesTable getSimTable() {
		return simTable;
	}

	/**
	 * This method close all the objects connected to this CliqueConstructor
	 * 
	 * @throws SQLException
	 */
	public void close() throws SQLException {
		if (psQueryGetBestIn != null)
			psQueryGetBestIn.close();
		if (psQueryPoleSim != null)
			psQueryPoleSim.close();

		connection.close();
	}
}
