package clustering.poboc.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;

import similarity.matrix.SimilaritiesTable;

/**
 * This class encapsulates the Clusters database table, and all methods required
 * to fill it up starting from the Poles and Similarities tables.
 * 
 * The Clusters table contains the result of the PoBOC clustering algorithm, it
 * contains for every cluster every objects belonging to it. 
 * The Clusters table is very similar to the Poles table but it allows duplicated
 * tuples, i.e., an object can belong to more than one cluster.
 * 
 * @author Luca Bernardini, Matteo Collina
 *
 */
public class ClustersTable {

	// Tabella risultato dell'algoritmo di Clustering.
	// Infatti contiene per ogni Cluster tutti gli oggetti Id a lui appartenenti
	// L'interfaccia della tabella è simile a quella della tabella Poles con una differenza
	// centrale ai fini del lavoro: Id può essere ripetuto nella tabella, perchè 
	// un oggetto può far parte di più cluster (gruppi)
	/**
	 * The query that creates the Connections table.
	 */
	public static final String createClustersTable = "CREATE TABLE Clusters ("
			+ "Id int NOT NULL," + // un oggetto può far parte di più cluster (gruppi)
			"Cluster int NOT NULL)";

	// Query inserisce una tupla in Clusters
	/**
	 * This query insert a tuple in Clusters.
	 */
	public static final String insertCluster = "INSERT INTO Clusters (Id,Cluster) VALUES (?,?)";
	
	
	// Query che restituisce tutti i Cluster di cui fa parte un oggetto
	/**
	 * This query returns all the objects that belongs to a Cluster.
	 */
	public static final String getObjectClusters = "SELECT Cluster FROM Clusters WHERE Id = ?";
	
	/**
	 * This query returns all the discovered Clusters.
	 */
	public static final String getClusters = "SELECT DISTINCT Cluster FROM Clusters";
	
	// Query che restituisce le similarità fra i cluster input1? e input2?
	public static final String getSimClusters = "" +
			"SELECT AVG(S.Sim) " +
			"FROM Similarities S JOIN Clusters C1 ON (S.Id_i = C1.Id) " +
			"JOIN Clusters C2 ON (S.Id_j = C2.Id) " +
			"WHERE C1.Cluster = ? AND C2.Cluster = ? ";
	
	private static final String getNumberClustersObjects = "SELECT COUNT(*) FROM Clusters WHERE Cluster = ?";
	
	/** 
	 *  The connection through we interact with the database.
	 */
	private Connection connection = null;
	
	private PolesTable polesTable = null;
	
	
	// Questi attributi sono invece le PreparateStatement che vengono invocate dai metodi che le incapsulano.
	// Grazie a questo approccio la costruzione della PreparateStatement avviene una volta sola (nel metodo get),
	// garantendo un risparmio evidente in sede di chiamata multipla delle query.
	private PreparedStatement psInsertCluster;
	private PreparedStatement psGetObjectClusters;
	private PreparedStatement psGetClusters;
	private PreparedStatement psGetSimClusters;
	private PreparedStatement psGetNumberClustersObjects;
	
	/**
	 * @param connection
	 * @throws SQLException 
	 */
	public ClustersTable(Connection connection, PolesTable polesTable) throws SQLException {
		this(connection,true);
		this.polesTable= polesTable;
	}
	
	/**
	 * @param connection
	 * @throws SQLException 
	 */
	public ClustersTable(Connection connection, boolean create) throws SQLException {
		super();
		this.connection = connection;
		if (create) {
			Statement stmt = connection.createStatement();
			stmt.executeUpdate(ClustersTable.createClustersTable);
			// Indice secondary unclustered single-attribute su Id
			stmt.executeUpdate("CREATE INDEX IdCLUIDX ON Clusters(Id)");
			// Indice secondary unclustered single-attribute su Cluster
			stmt
					.executeUpdate("CREATE INDEX ClusterCLUIDX ON Clusters(Cluster)");
			stmt.close();
		}
	}

	/**
	 * This method asks for all the objects in the Similarities table and for every of them
	 * it calls assignObjectToClusters.
	 * @param similaritiesTable
	 * @throws SQLException
	 */
	public void doClustering(SimilaritiesTable similaritiesTable)
			throws SQLException {
		ResultSet objectsIds = similaritiesTable.getObjectsIds();
		int id;

		int i = 0;
		
		connection.setAutoCommit(false);
		
		while (objectsIds.next()) {
			id = objectsIds.getInt(1);
			this.assignObjectToClusters(id);
			if(++i % 100 == 0)
				System.out.println("Assigned " + i + " objects");
		}
		
		connection.commit();
		connection.setAutoCommit(true);
		objectsIds.close();
	}

	// ASSIGN
	// pass, current, future sono dei cluster: la mia finestra scorrevole è di
	// dimensione 3
	/**
	 * This method implement the ASSIGN function as described in the paper.
	 * 
	 * @param id
	 *            The object to assign to the clusters.
	 */
	protected void assignObjectToClusters(int id) throws SQLException {
		ResultSet orderedPoles = polesTable.getPoleSimilarities(id);
		if (!orderedPoles.next())
			throw new RuntimeException("There aren't any clusters!");

		// l'oggetto appartiene ad almeno un cluster quindi è associato di
		// default al primo della lista in ordine di similarità
		int current = orderedPoles.getInt(1);
		float poleSimCurrent = orderedPoles.getFloat(2);
		this.insertCluster(id, current);

		// l'oggetto appartiene al current cluster se la similarità con esso
		// è maggiore della media della similarità che questo oggetto ha con
		// pass e future

		int future;
		float poleSimPass = poleSimCurrent;
		float poleSimFuture, threshold = 0;

		if (!orderedPoles.next()) {
			// If we're here its because we haven't fully built the similarity matrix
			// so this object is similar to only one cluster.
			// throw new RuntimeException("There is only one cluster!");
			orderedPoles.close();
			return;
		}
		current = orderedPoles.getInt(1);
		poleSimCurrent = orderedPoles.getFloat(2);

		boolean addedToCluster = true;
		
		while (true) {
			if (!orderedPoles.next())
				break;
			future = orderedPoles.getInt(1);
			poleSimFuture = orderedPoles.getFloat(2);

			threshold = (poleSimPass + poleSimFuture) / 2;

			if (poleSimCurrent >= threshold) {
				this.insertCluster(id, current);
			} else {
				addedToCluster = false;
				break;
			}

			poleSimPass = poleSimCurrent;
			current = future;
			poleSimCurrent = poleSimFuture;
		}
		// Se addedToCluster qui è true, significa che siamo qui grazie alla
		// prima break.
		// In questo caso in Current c'è l'ultimo polo e in Pass il penultimo.
		// L'oggetto fa già parte del Pass e se la similarità con il Current è
		// uguale
		// a quella che ha con il Pass allora l'oggetto farà parte anche del
		// Current

		if (addedToCluster && poleSimPass == poleSimCurrent) {
			this.insertCluster(id, current);
		}

		orderedPoles.close();

		System.runFinalization();
		System.gc();
	}
	
	/**
	 * This method is used to insert a tuple in the Clusters table.
	 * 
	 * @param id
	 *            The identifier of the object.
	 * @param cluster
	 *            The identifier of the cluster.
	 * @throws SQLException
	 */
	public void insertCluster(int id, int cluster) throws SQLException {
		PreparedStatement ps = this.getPsInsertCluster();
		ps.setInt(1, id);
		ps.setInt(2, cluster);
		ps.executeUpdate();
	}

	/**
	 * This methods returns an array which contains all the clusters which the
	 * given id is belonging.
	 * 
	 * @param id
	 * @return
	 * @throws SQLException
	 */
	public int[] getClusters(int id) throws SQLException {
		PreparedStatement ps = this.getPsGetObjectClusters();
		ps.setInt(1, id);
		ResultSet rs = ps.executeQuery();

		ArrayList<Integer> idClusters = new ArrayList<Integer>();
		while (rs.next()) {
			idClusters.add(rs.getInt(1));
		}
		rs.close();

		int[] result = new int[idClusters.size()];
		for (int i = 0; i < result.length; i++) {
			result[i] = idClusters.get(i);
		}
		return result;
	}

	/**
	 * @return All the discovered clusters.
	 * @throws SQLException
	 */
	public int[] getClusters() throws SQLException {
		PreparedStatement ps = this.getPsGetClusters();
		ResultSet rs = ps.executeQuery();

		ArrayList<Integer> idClusters = new ArrayList<Integer>();
		while (rs.next()) {
			idClusters.add(rs.getInt(1));
		}
		rs.close();

		int[] result = new int[idClusters.size()];
		for (int i = 0; i < result.length; i++) {
			result[i] = idClusters.get(i);
		}
		return result;
	}

	public float getSimClusters(int C1, int C2) throws SQLException {
		PreparedStatement ps = this.getPsGetSimClusters();
		ps.setInt(1, C1);
		ps.setInt(2, C2);
		ResultSet rs = ps.executeQuery();
		rs.next();
		return rs.getFloat(1);
	}

	public boolean thereIsOverlap() throws SQLException {
		Statement stmt = this.connection.createStatement();
		ResultSet rs = stmt
				.executeQuery("SELECT Id, COUNT(Id) FROM Clusters GROUP BY Id HAVING COUNT(Id) > 1 ");
		if (rs.next()) {
			stmt.close();
			return true;
		}
		stmt.close();
		return false;
	}

	private PreparedStatement getPsInsertCluster() throws SQLException {
		if (this.psInsertCluster == null)
			psInsertCluster = connection
					.prepareStatement(ClustersTable.insertCluster);
		return psInsertCluster;
	}

	private PreparedStatement getPsGetObjectClusters() throws SQLException {
		if (this.psGetObjectClusters == null)
			psGetObjectClusters = connection
					.prepareStatement(ClustersTable.getObjectClusters);
		return psGetObjectClusters;
	}

	private PreparedStatement getPsGetClusters() throws SQLException {
		if (this.psGetClusters == null)
			psGetClusters = connection
					.prepareStatement(ClustersTable.getClusters);
		return psGetClusters;
	}

	private PreparedStatement getPsGetSimClusters() throws SQLException {
		if (this.psGetSimClusters == null)
			psGetSimClusters = connection
					.prepareStatement(ClustersTable.getSimClusters);
		return psGetSimClusters;
	}

	public void printTable() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt
				.executeQuery("SELECT * FROM Clusters ORDER BY Cluster");
		System.out.println("Id ---  Cluster ");
		while (rs.next()) {
			System.out.println("" + rs.getInt(1) + " -- " + rs.getInt(2));
		}
		stmt.close();
	}

	/**
	 * This method close all the objects connected to this ClustersTable
	 * 
	 * @throws SQLException
	 */
	public void close() throws SQLException {
		if (psInsertCluster != null)
			psInsertCluster.close();
		if (psGetObjectClusters != null)
			psGetObjectClusters.close();
		if (psGetClusters != null)
			psGetClusters.close();
		if (psGetSimClusters != null)
			psGetSimClusters.close();
		if (psGetNumberClustersObjects != null)
			psGetNumberClustersObjects.close();
		
		this.connection.close();
	}

	public PolesTable getPolesTable() {
		return polesTable;
	}

	public void setPolesTable(PolesTable polesTable) {
		this.polesTable = polesTable;
	}

	public int getNumberOfClustersFound() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT DISTINCT Cluster FROM Clusters");
		int result = 0;
		while(rs.next())
			result++;	
		stmt.close();
		return result;
	}

	public int getNumberOfElementsInCluster(int clusterIndex) throws SQLException {
		PreparedStatement ps = this.getPsGetNumberClustersObjects();
		ps.setInt(1, clusterIndex);
		ResultSet rs = ps.executeQuery();
		rs.next();
		int result = rs.getInt(1);		
		return result;
	}

	private PreparedStatement getPsGetNumberClustersObjects() throws SQLException {
		if (psGetNumberClustersObjects==null)
			psGetNumberClustersObjects = connection.prepareStatement(getNumberClustersObjects);
		return psGetNumberClustersObjects;
	}

	public int getSize() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT COUNT(*) FROM Clusters");
		rs.next();
		int result = rs.getInt(1);		
		stmt.close();
		return result;
	}

}
