package clustering.poboc.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import similarity.matrix.AveragesTable;
import similarity.matrix.SimilaritiesTable;

// Tabella che contiene le informazioni sul grafo, ovvero tutte le connessioni
// (in modo ridondante nel senso che se c'è (1,2) c'è anche (2,1)).
// Ci serve in fase di individuazione delle cricche per sapere quali sono i vertici con almeno una connessione 
/**
 * This table contains the Similarity Graph as stated in the paper.
 * It's redundant like {@link SimilaritiesTable}.
 * 
 * This table its needed when we need to find the cliques and we need
 * to know what vertexes are connected.
 * 
 * @author Luca Bernardini, Matteo Collina
 *
 */
public class ConnectionsTable {

	/**
	 * The query that creates the Connections table.
	 */
	public static final String createConnectionsTable = "CREATE TABLE Connections ("
			+ "Id_i int NOT NULL,"
			+ "Id_j int NOT NULL,"
			+ "Sim real NOT NULL,"
			+ "PRIMARY KEY (Id_i,Id_j))";

	// Vista della tabella Connections, che rappresenti la relazione: gli oggetti appartenenti a questa vista
	// hanno almeno una connessione.Id_i
	// Quindi non c'è ridondanza!
	
	/**
	 * This query creates the Nodes view which are all the Nodes with at least one connection.
	 */
	public static final String createNodesView = "CREATE VIEW Nodes(Id) AS "
			+ "SELECT DISTINCT C.Id_i " + "FROM Connections C ";

	/**
	 * This query automatically fill the Connections table from the Similarities table
	 * and the Averages table.  
	 */
	public static final String insertConnection = "INSERT INTO Connections (Id_i,Id_j,Sim) " +
		"SELECT S.Id_i, S.Id_j, S.Sim " +
//		"FROM Similarities S JOIN Averages A ON (S.Id_i = A.Id) " +
//		"WHERE S.Sim >= ( " +
//			"SELECT A2.RowAvg " +
//			"FROM Averages A2 " +
//			"WHERE A2.Id = S.Id_i ) " +
//		"AND S.Sim >= (" +
//			"SELECT A2.RowAvg " +
//			"FROM Averages A2 " +
//			"WHERE A2.Id = S.Id_j )";
	
		"FROM Similarities S JOIN Averages A1 ON (S.Id_i = A1.Id) " +
				"JOIN Averages A2 ON (S.Id_j = A2.Id) " +
		"WHERE S.Sim >= A1.RowAvg AND S.Sim >= A2.RowAvg ";
		
	
	// Query che interroga Connections e restituisce tutti i nodi Id_j connessi col nodo dato in input:
	/**
	 * This query returns all the nodes connected to the nodes given in input.
	 */
	public static final String getConnections = "SELECT Id_j FROM Connections WHERE Id_i= ? ";
	
	/** 
	 *  The connection through we interact with the database.
	 */
	private Connection connection = null;

	private PreparedStatement psGetConnection;
	
	/**
	 * @param connection
	 * @throws SQLException 
	 */
	public ConnectionsTable(Connection connection) throws SQLException {
		super();
		this.connection = connection;
		Statement stmt = connection.createStatement();
		stmt.executeUpdate(ConnectionsTable.createConnectionsTable);
		// View
		stmt.executeUpdate(ConnectionsTable.createNodesView);
		stmt.close();
	}

	/** 
	 *  This method automatically fill the connections table
	 *  
	 *  @param Similarity matrix , AveragesTable averages
	 * 	@throws SQLException 
	 *  @throws IllegalArgumentException if matrix or averages are null.
	 */
	public void fillConnectionTable (SimilaritiesTable matrix, AveragesTable averages) throws SQLException {

		if ( matrix == null || averages == null ) throw new NullPointerException();
		Statement stmt = connection.createStatement();
		stmt.executeUpdate(ConnectionsTable.insertConnection);
		

		// Indice primary clustered multi-attributo su (Id_i,Id_j)
		stmt.executeUpdate("CREATE UNIQUE INDEX IdiIdjCONNIDX ON Connections(Id_i,Id_j)");
		stmt.executeUpdate("CREATE INDEX IdiSimCONNIDX ON Similarities(Id_i,Sim DESC)");
		stmt.close();

	}

	/** 
	 * 
	 *  @param id_i
	 *  @return the stored id_j if connection (id_i,id_j) exists
	 *  @throws SQLException if the table doesn't contains a (id_i,id_j)
	 */
	public ResultSet getConnection(int id_i) throws SQLException {
		
		PreparedStatement ps = this.getPsGetConnection();
		ps.setInt(1, id_i);
		ResultSet result = ps.executeQuery();
		return result;
	}
	
	//
	//  GET delle PreparedStatement: quando le invoco, se non è ancora stata creata, 
	//  viene creata la PreparedStatment passando la stringa relativa
	//
	
	private PreparedStatement getPsGetConnection() throws SQLException {
		if(psGetConnection == null)
			psGetConnection = connection.prepareStatement(ConnectionsTable.getConnections);
		return psGetConnection;
	}
	
	public void printTable() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT * FROM Connections");
		System.out.println("Id_i ---  Id_j --- Sim ");
		while (rs.next()) {
			System.out.println(rs.getInt(1) + " -- " + rs.getInt(2) + " -- "
					+ rs.getFloat(3));
		}
		stmt.close();
	}


	public int getSize() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT COUNT(*) FROM Connections");
		rs.next();
		int result = rs.getInt(1);		
		stmt.close();
		return result;
	}
	
	/**
	 * This method close all the objects connected to this SimilaritiesTable
	 * 
	 * @throws SQLException
	 */
	public void close() throws SQLException {
		if (psGetConnection != null)
			psGetConnection.close();

		connection.close();
	}


}
