package clustering.poboc.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

// Tabella che mantiene per ogni Polo (cricca) tutti gli oggetti a lui appartenenti.
// Per avere gli oggetti di un polo specifico si fa una SELECT su Poles.Id con WHERE sul valore di Pole
/**
 * This Table mantains for every poles all the objects that it contains.
 *  
 * @author Luca Bernardini, Matteo Collina
 *
 */
public class PolesTable {

	/**
	 * The query that creates the Connections table.
	 */
	public static final String createPolesTable = "CREATE TABLE Poles ("
			+ "Id int NOT NULL PRIMARY KEY," + // un oggetto fa parte di una sola cricca
			"Pole int NOT NULL)";

	// Vista per avere la cardinalità di un polo
	/**
	 * This query creates the PoleCounts view that's used to know the cardinality of a Pole.
	 */
	public static final String createPoleCountsView = "CREATE VIEW PoleCounts (Pole,PoleCount) AS "
			+ "SELECT P.Pole, COUNT(*) "
			+ "FROM Poles P "
			+ "WHERE Pole = P.Pole "
			+ "GROUP BY P.Pole ";

	// Vista per avere la similarità fra un punto (fra tutti gli N del dataset) e tutti i Poli (sono K), calcolata come media delle similarità tra
	// il punto in questione e tutti gli elementi appartenenti al polo. La vista avrà N (righe)x K (colonne), ma sarà gestita dal DBMS
	/**
	 * This query creates the PoleSimilarities View that allows to know the similarity between an object and all poles.
	 * Its calculated as the average of all the similarities between that object and all object in a pole.
	 */
	public static final String createPoleSimilaritiesView = "CREATE VIEW PoleSimilarities (Id,Pole,PoleSim) AS "
			+ "SELECT S.Id_j, P.Pole, AVG(S.Sim) AS PoleSim "
			+ "FROM Similarities S JOIN Poles P ON (P.Id = S.Id_i) "
			+ "GROUP BY S.Id_j, P.Pole";
	
	// Query inserisce una tupla in Poles : 
	/**
	 * This query insert a tuple in Poles.
	 */
	public static final String insertPole = "INSERT INTO Poles (Id,Pole) VALUES (?,?)";
	
	// Query che elimina una tupla in Poles : 
	/**
	 * This query delete a Pole (delete all Pole's tuple of a specified Pole)
	 */
	public static final String deletePole = "DELETE FROM Poles WHERE Pole = ?";

	// Query che interroga Poles e restituisce tutti i nodi Id che fanno parte del polo
	/**
	 * This query returns all the objects that belongs to a pole.
	 */
	public static final String getPoleObjects = "SELECT Id FROM Poles WHERE Pole= ? ";
	
	// Query che interroga PoleCounts e restituisce la dimensione del polo richiesto
	/**
	 * 
	 */
	public static final String getNumberPoleObjects = "SELECT COUNT(*) FROM Poles WHERE Pole = ?";
	
	// Query che interroga la view PoleSimilarities e dato in input un oggetto restituisce le K similarità coi K poli
	private static final String getPoleSimilarities = "SELECT Pole,PoleSim " +
													  "FROM PoleSimilarities " +
													  "WHERE Id = ? " +
													  "ORDER BY PoleSim DESC";
	
	/** 
	 *  The connection through we interact with the database.
	 */
	private Connection connection = null;
	
	
	// Questi attributi sono invece le PreparateStatement che vengono invocate dai metodi che le incapsulano.
	// Grazie a questo approccio la costruzione della PreparateStatement avviene una volta sola (nel metodo get),
	// garantendo un risparmio evidente in sede di chiamata multipla delle query.
	private PreparedStatement psInsertPole;
	private PreparedStatement psDeletePole;
	private PreparedStatement psGetPoleObjects;
	private PreparedStatement psGetPoleSimilarity;
	private PreparedStatement psGetNumberPoleObjects;
	
	/**
	 * @param connection
	 * @throws SQLException 
	 */
	public PolesTable(Connection connection) throws SQLException {
		super();
		this.connection = connection;
		Statement stmt = connection.createStatement();
		stmt.executeUpdate(PolesTable.createPolesTable);
		// Indice primary clustered single-attribute su Id
		stmt.executeUpdate("CREATE UNIQUE INDEX IdPOLIDX ON Poles(Id)");
		// Indice secondary unclustered single-attribute su Pole
		stmt.executeUpdate("CREATE INDEX PolePOLIDX ON Poles(Pole)");
		// View
		stmt.executeUpdate(PolesTable.createPoleCountsView);
		stmt.executeUpdate(PolesTable.createPoleSimilaritiesView);
		
		stmt.close();
	}
	
	
	/** 
	 *  This method insert an element in pole specified
	 *  
	 *  @param id e Pole
	 * 	@throws SQLException 
	 */
	public void insertPole (int id, int pole) throws SQLException {

		PreparedStatement ps = this.getPsInsertPole();
		ps.setInt(1, id);
		ps.setInt(2, pole);
		ps.executeUpdate();
		
	}
	
	/** 
	 *  This method delete a Pole (delete all Pole's tuple of a specified Pole)
	 *  
	 *  @param id e Pole
	 * 	@throws SQLException 
	 */
	public void deletePole (int pole) throws SQLException {

		PreparedStatement ps = this.getPsDeletePole();
		ps.setInt(1, pole);
		ps.executeUpdate();
		
	}
	
	/** 
	 * 
	 *  @param pole
	 *  @return objects id which are in the pole
	 *  @throws SQLException if the table doesn't contains a pole
	 */
	public ResultSet getPoleObjects(int pole) throws SQLException {
		
		PreparedStatement ps = this.getPsGetPoleObjects();
		ps.setInt(1, pole);
		return ps.executeQuery();
	}
	
	/** 
	 * 	Returns a ResultSet which contains (Pole,PoleSim), where PoleSim is the similarity
	 *  between pole and the given id.
	 * 	Pole is an INT, while PoleSim is an REAL
	 * 
	 *  @param id
	 *  @return ResultSet
	 *  @throws SQLException if the table doesn't contains a pole
	 */
	
	public ResultSet getPoleSimilarities(int id) throws SQLException{
		
		PreparedStatement ps = this.getPsGetPoleSimilarity();
		ps.setInt(1, id);
		return ps.executeQuery();
	}
	
	//
	//  GET delle PreparedStatement: quando le invoco, se non è ancora stata creata, 
	//  viene creata la PreparedStatment passando la stringa relativa
	//
	
	private PreparedStatement getPsInsertPole() throws SQLException {
		if(psInsertPole == null)
			psInsertPole = connection.prepareStatement(PolesTable.insertPole);
		return psInsertPole;
	}
	
	private PreparedStatement getPsDeletePole() throws SQLException {
		if(psDeletePole == null)
			psDeletePole = connection.prepareStatement(PolesTable.deletePole);
		return psDeletePole;
	}

	private PreparedStatement getPsGetPoleObjects() throws SQLException {
		if(psGetPoleObjects == null)
			psGetPoleObjects = connection.prepareStatement(PolesTable.getPoleObjects);
		return psGetPoleObjects;
	}
	
	private PreparedStatement getPsGetPoleSimilarity() throws SQLException {
		if(psGetPoleSimilarity == null)
			psGetPoleSimilarity = connection.prepareStatement(PolesTable.getPoleSimilarities);
		return psGetPoleSimilarity;
	}
		
	public PreparedStatement getPsGetNumberPoleObjects() throws SQLException {
		if(psGetNumberPoleObjects == null)
			psGetNumberPoleObjects = connection.prepareStatement(PolesTable.getNumberPoleObjects);
		return psGetNumberPoleObjects;
	}
	
	public void printTable() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT * FROM Poles");
		System.out.println("Id ---  Pole ");
		while (rs.next()) {
			System.out.println("" + rs.getInt(1) + " -- " + rs.getInt(2));
		}
		stmt.close();
	}

	/**
	 * This method close all the objects connected to this PolesTable
	 * 
	 * @throws SQLException
	 */
	public void close() throws SQLException {
		if (psInsertPole != null)
			psInsertPole.close();
		if (psDeletePole != null)
			psDeletePole.close();
		if (psGetPoleObjects != null)
			psGetPoleObjects.close();
		if (psGetPoleSimilarity != null)
			psGetPoleSimilarity.close();
		if (psGetNumberPoleObjects != null)
			psGetNumberPoleObjects.close();

		connection.close();
	}


	public Connection getConnection() {
		return connection;
	}


	public void setConnection(Connection connection) {
		this.connection = connection;
	}


	public int getNumberPolesFound() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT DISTINCT Pole FROM Poles");
		int result = 0;
		while(rs.next())
			result++;
		stmt.close();
		return result;
	}

	
	public int getNumberElementPole(int poleIndex) throws SQLException {
		PreparedStatement ps = this.getPsGetNumberPoleObjects();
		ps.setInt(1, poleIndex);
		ResultSet rs = ps.executeQuery();
		rs.next();
		int result = rs.getInt(1);		
		return result;
	}


	public int getSize() throws SQLException {
		Statement stmt = connection.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT COUNT(*) FROM Poles");
		rs.next();
		int result = rs.getInt(1);		
		stmt.close();
		return result;
	}

	
}
